#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <libdpfp/dpfp.h>

/* Change this to whatever your daemon is called */
#define DAEMON_NAME "dpfp-daemon"

/* Change this to the user under which to run */
#define RUN_AS_USER "root"

#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1

#ifndef LINE_MAX
#define LINE_MAX 255
#endif

#define TV_TO_DOUBLE( tv ) ( tv.tv_sec + ( tv.tv_usec / 1000000.0 ) )

static void LOG( const char *x, ... ) {
  char buf[LINE_MAX];
  va_list a;
  va_start( a, x );
  vsnprintf( buf, sizeof( buf ), x, a );
  va_end( a );
  syslog( LOG_AUTH|LOG_INFO, DAEMON_NAME ": %s", buf );
}

int get_dpfp()
{
  int val;
  double t1, t2;

  struct timeval tv;

  struct dpfp_dev *dev;
  struct dpfp_fprint *base_img = dpfp_fprint_alloc();
  struct dpfp_fprint *fp = dpfp_fprint_alloc();
  struct dpfp_fprint *mask = dpfp_fprint_alloc();
  struct dpfp_ffield *direction = dpfp_ffield_alloc();
  struct dpfp_ffield *frequency = dpfp_ffield_alloc();

  gettimeofday( &tv, NULL );
  t1 = TV_TO_DOUBLE( tv );

  dpfp_init();
  dev = dpfp_open();

  if( dev == NULL ) {
    LOG( "libdpfp failed to open" );
    val = 1;
  }

  /* Capture base image. We assume the finger is away from the sensor at
   * this point... */
  if( dpfp_set_mode( dev, DPFP_MODE_SEND_FINGER ) < 0 ) {
    LOG( "libdpfp failed to set mode" );
    val = 1;    
  }

  if( dpfp_capture_fprint( dev, base_img ) < 0 ) {
    LOG( "libdpfp failed on simple_get_fingerprint" );
    val = 1;
  }

  if( dpfp_simple_await_finger_on( dev ) < 0 ) {
    LOG( "libdpfp failed on await_finger_on" );
    val = 1;
  }
  /* Capture fingerprint */
  if( dpfp_set_mode( dev, DPFP_MODE_SEND_FINGER ) < 0 ) {
    LOG( "libdpfp failed on set_mode" );
    val = 1;
  }
  if( dpfp_capture_fprint( dev, fp ) < 0 ) {
    LOG( "failed on simple_get_fingerprint" );
    val = 1;
  }
  dpfp_close( dev );

  /* Basic enhancements: subtract base image, flip to correct orientation */
  dpfp_fprint_subtract( fp, base_img );
  dpfp_fprint_flip_v( fp );
  dpfp_fprint_flip_h( fp );
  if( dpfp_fprint_write_to_file( fp, "finger.pgm" ) < 0 ) {
    LOG( "libdpfp failed on write_fingerprint_to_file" );
    val = 1;
  }
  /* More advanced enhancements */
  dpfp_fprint_soften_mean( fp, 3 );
  dpfp_fprint_get_direction( fp, direction, 7, 8 );
  dpfp_fprint_get_frequency( fp, direction, frequency );
  dpfp_fprint_get_mask( fp, direction, frequency, mask );
  dpfp_fprint_enhance_gabor( fp, direction, frequency, mask, 4.0 );
  dpfp_fprint_binarize( fp, 0x80 );
  if( dpfp_fprint_write_to_file( fp, "finger_enhanced.pgm" ) < 0 ) {
    LOG( "failed on write_fingerprint_to_file" );
    val = 1;
  }
  dpfp_fprint_thin(fp);
  gettimeofday(&tv, NULL);
  t2 = TV_TO_DOUBLE(tv);
  printf( "enhancements took %.6lf seconds in total\n", t2 - t1 );
  if( dpfp_fprint_write_to_file( fp, "finger_thinned.pgm" ) < 0 ) {
    LOG( "libdpfp failed on write_fingerprint_to_file" );
    val = 1;
  }

  return val;
}

static void child_handler( int signum )
{
  switch(signum) {
  case SIGALRM: exit( EXIT_FAILURE ); break;
  case SIGUSR1: exit( EXIT_SUCCESS ); break;
  case SIGCHLD: exit( EXIT_FAILURE ); break;
  }
}

int daemonize( const char *lockfile )
{
  pid_t pid, sid, parent;
  int lfp = -1;
  
  /* already a daemon */
  if ( getppid() == 1 ) return;
  
  /* Create the lock file as the current user */
  if ( lockfile && lockfile[0] ) {
    lfp = open( lockfile, O_RDWR|O_CREAT, 0640 );
    if ( lfp < 0 ) {
      LOG( "unable to create lock file %s, code=%d ( %s )",
	   lockfile, errno, strerror( errno ) );
      exit( EXIT_FAILURE );
    }
  }
  
  /* Drop user if there is one, and we were run as root */
  if ( getuid() == 0 || geteuid() == 0 ) {
    struct passwd *pw = getpwnam(RUN_AS_USER);
    if ( pw ) {
      LOG( "setting user to " RUN_AS_USER );
      setuid( pw->pw_uid );
    }
  }
  
  /* Trap signals that we expect to recieve */
  signal( SIGCHLD, child_handler );
  signal( SIGUSR1, child_handler );
  signal( SIGALRM, child_handler );
  
  /* Fork off the parent process */
  pid = fork();
  if( pid < 0 ) {
    LOG( "unable to fork daemon, code=%d ( %s )",
	 errno, strerror( errno ) );
    exit( EXIT_FAILURE );
  }
  /* If we got a good PID, then we can exit the parent process. */
  if( pid > 0 ) {
    
    /* Wait for confirmation from the child via SIGTERM or SIGCHLD, or
       for two seconds to elapse (SIGALRM).  pause() should not return. */
    alarm( 2 );
    pause();
    
    exit( EXIT_FAILURE );
  }
  
  /* At this point we are executing as the child process */
  parent = getppid();
  
  /* Cancel certain signals */
  signal( SIGCHLD,SIG_DFL ); /* A child process dies */
  signal( SIGTSTP,SIG_IGN ); /* Various TTY signals */
  signal( SIGTTOU,SIG_IGN );
  signal( SIGTTIN,SIG_IGN );
  signal( SIGHUP, SIG_IGN ); /* Ignore hangup signal */
  signal( SIGTERM,SIG_DFL ); /* Die on SIGTERM */
  
  /* Change the file mode mask */
  umask(0);
  
  /* Create a new SID for the child process */
  sid = setsid();
  if( sid < 0 ) {
    LOG( "unable to create a new session, code %d ( %s )",
	 errno, strerror( errno ) );
    exit( EXIT_FAILURE );
  }
  
  /* Change the current working directory.  This prevents the current
     directory from being locked; hence not being able to remove it. */
  if( ( chdir( "/" ) ) < 0 ) {
    LOG( "unable to change directory to %s, code %d ( %s )",
	 "/", errno, strerror( errno ) );
    exit( EXIT_FAILURE );
  }

  /* Redirect standard files to /dev/null */
  freopen( "/dev/null", "r", stdin );
  freopen( "/dev/null", "w", stdout );
  freopen( "/dev/null", "w", stderr );
  
  /* Tell the parent process that we are A-okay */
  kill( parent, SIGUSR1 );
}

int main( int argc, char *argv[] ) {

  /* One may wish to process command line arguments here */
  
  /* Daemonize */
  while( daemonize( "/var/lock/subsys/" DAEMON_NAME ) != 1 ) {
    /* Now we are a daemon -- do the work for which we were paid */
    while( get_dpfp() != 1 ) {
      LOG( "scanning....\n" );
    }
  }
  
  return 0;
}
