/*
 * libdpfp example to capture a fingerprint and enhance it, then write it
 * to a PGM file
 * 
 *    Copyright (C) 2006 Daniel Drake <dsd@gentoo.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Compile:
 * gcc -o libdpfp libdpfp.c -ldpfp
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/unistd.h>
#include <sys/select.h>
#include <sys/time.h>

#include <libdpfp/dpfp.h>

#define TV_TO_DOUBLE( tv ) ( tv.tv_sec + ( tv.tv_usec / 1000000.0 ) )

volatile sig_atomic_t kg = 1;

void CatchTerm( int sig )
{
  kg = 0;
  signal( sig, CatchTerm );
}

int GetFP()
{
  int val;
  double t1, t2;

  struct timeval tv;
  struct dpfp_fprint *mask = dpfp_fprint_alloc();
  struct dpfp_ffield *direction = dpfp_ffield_alloc();
  struct dpfp_ffield *frequency = dpfp_ffield_alloc();
  struct dpfp_dev *dev;
  struct dpfp_fprint *base_img = dpfp_fprint_alloc();
  struct dpfp_fprint *fp = dpfp_fprint_alloc();

  gettimeofday( &tv, NULL );
  t1 = TV_TO_DOUBLE( tv );

  dpfp_init();
  dev = dpfp_open();
  if( dev == NULL ) {
    printf( "libdpfp failed to open" );
  }
  /* Capture base image. We assume the finger is away from the sensor at
   * this point... */
  if( dpfp_set_mode( dev, DPFP_MODE_SEND_FINGER ) < 0 ) {
    printf( "libdpfp failed to set mode" );
    
  }
  if( dpfp_capture_fprint( dev, base_img ) < 0 ) {
    printf( "libdpfp failed on simple_get_fingerprint" );
  }
  if( dpfp_simple_await_finger_on( dev ) < 0 ) {
    printf( "libdpfp failed on await_finger_on" );
  }
  /* Capture fingerprint */
  if( dpfp_set_mode( dev, DPFP_MODE_SEND_FINGER ) < 0 ) {
    printf( "libdpfp failed on set_mode" );
  }
  if( dpfp_capture_fprint( dev, fp ) < 0 ) {
    printf( "failed on simple_get_fingerprint" );
  }
  dpfp_close( dev );

  /* Basic enhancements: subtract base image, flip to correct orientation */
  dpfp_fprint_subtract( fp, base_img );
  dpfp_fprint_flip_v( fp );
  dpfp_fprint_flip_h( fp );
  if( dpfp_fprint_write_to_file( fp, "finger.pgm" ) < 0 ) {
    printf( "libdpfp failed on write_fingerprint_to_file" );
    val = 1;
  }
  /* More advanced enhancements */
  dpfp_fprint_soften_mean( fp, 3 );
  dpfp_fprint_get_direction( fp, direction, 7, 8 );
  dpfp_fprint_get_frequency( fp, direction, frequency );
  dpfp_fprint_get_mask( fp, direction, frequency, mask );
  dpfp_fprint_enhance_gabor( fp, direction, frequency, mask, 4.0 );
  dpfp_fprint_binarize( fp, 0x80 );
  if( dpfp_fprint_write_to_file( fp, "finger_enhanced.pgm" ) < 0 ) {
    printf( "failed on write_fingerprint_to_file" );
    val = 1;
  }
  dpfp_fprint_thin(fp);
  gettimeofday(&tv, NULL);
  t2 = TV_TO_DOUBLE(tv);
  printf( "enhancements took %.6lf seconds in total\n", t2 - t1 );
  if( dpfp_fprint_write_to_file( fp, "finger_thinned.pgm" ) < 0 ) {
    printf( "libdpfp failed on write_fingerprint_to_file" );
    val = 1;
  }
}

int main( void )
{
  signal( SIGALRM, CatchTerm );
  alarm( 0 );
  while( kg ) {
    GetFP();
  }
  return EXIT_SUCCESS;
}

