/*
   pam_libdpfp - Provides a PAM-compliant interface to use in biometrically authenticating users

   Currently supports remote databases such as OpenLDAP/Active Directory, MySQL and Kerberized
   authentication.

   Copyright (C) 2007 Jason L. Gerfen <jason.gerfen@gmail.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#define _GNU_SOURCE
#define PAM_SM_AUTH
#define PAM_SM_PASSWORD

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <security/_pam_macros.h>
#include <security/pam_modules.h>

#include <libdpfp/dpfp.h>

#ifdef INCLUDE_LIBMYSQL
#include <mysql.h>
#endif

#ifdef INCLUDE_LIBLDAP
#include <ldap.h>
#endif

#ifdef INCLUDE_LIBKRB5
#include <krb5.h>

#ifdef HAVE_SECURITY_PAM_APPL_H
#include <security/pam_appl.h>
#endif

#ifdef HAVE_SECURITY_PAM_MODULES_H
#include <security/pam_modules.h>
#endif

#ifdef HAVE_SECURITY_PAM_APPL_H
#include <security/pam_appl.h>
#endif

#ifdef HAVE_SECURITY_PAM_MODULES_H
#define PAM_SM_AUTH
#define PAM_SM_SESSION
#include <security/pam_modules.h>
#endif

#include "config.h"
#include "conv.h"
#include "items.h"
#include "prompter.h"
#include "init.h"
#include "initopts.h"
#include "items.h"
#include "log.h"
#include "options.h"
#include "prompter.h"
#include "sly.h"
#include "stash.h"
#include "tokens.h"
#include "userinfo.h"
#include "v5.h"
#include "v4.h"
#include "xstr.h"
#endif

#ifdef INCLUDE_DATABASES 
#define MODULE_NAME "pam_libdpfp+db"
#else
#define MODULE_NAME "pam_libdpfp"
#endif

#define TV_TO_DOUBLE(tv) (tv.tv_sec + (tv.tv_usec / 1000000.0))

#ifndef LINE_MAX
#define LINE_MAX 255
#endif

/* Logging macros */
static void
DEBUG( const char *x, ... ) {
  char buf[LINE_MAX];
  va_list a;
  va_start( a, x );
  vsnprintf( buf, sizeof( buf ), x, a );
  va_end( a );
  syslog( LOG_AUTH|LOG_DEBUG, MODULE_NAME ": %s", buf );
}
static void
NOTICE( const char *x, ... ) {
  char buf[LINE_MAX];
  va_list a;
  va_start( a, x );
  vsnprintf( buf, sizeof( buf ), x, a );
  va_end( a );
  syslog( LOG_AUTH|LOG_NOTICE, MODULE_NAME ": %s", buf );
}
static void
INFO( const char *x, ... ) {
  char buf[LINE_MAX];
  va_list a;
  va_start( a, x );
  vsnprintf( buf, sizeof( buf ), x, a );
  va_end( a );
  syslog( LOG_AUTH|LOG_INFO, MODULE_NAME ": %s", buf );
}
static void
WARN( const char *x, ... ) {
  char buf[LINE_MAX];
  va_list a;
  va_start( a, x );
  vsnprintf( buf, sizeof( buf ), x, a );
  va_end( a );
  syslog( LOG_AUTH|LOG_WARNING, MODULE_NAME ": %s", buf );
}
static void
CRIT( const char *x, ... ) {
  char buf[LINE_MAX];
  va_list a;
  va_start( a, x );
  vsnprintf( buf, sizeof( buf ), x, a );
  va_end( a );
  syslog( LOG_AUTH|LOG_CRIT, MODULE_NAME ": %s", buf );
}

#ifdef INCLUDE_DATABASES
#define CONFIG_FILE "/etc/pam_libdpfp.conf"

typedef struct myConfig * myConfigPtr;
struct myConfig {
  char usrname[45];
  char ldap_schema[4];
  char ldap_servers[255];
  char ldap_port[4];
  char ldap_bind_dn[255];
  char ldap_base_dn[255];
  char ldap_user[20];
  char ldap_pass[20];
  char ldap_tls[5];
  char mysql_servers[255];
  char mysql_username[45];
  char mysql_password[45];
  char mysql_database[45];
  char mysql_ssl[5];
  char id[45];
  char uid[20];
  char gid[20];
  char shell[20];
  char home[40];
} localConfig;

int ReadConfig()
{
  FILE *cfg;
  char line[1024];
  char *chk;
  myConfigPtr myConfig = &localConfig;
  
  if( ( cfg = fopen( CONFIG_FILE, "r" ) ) == NULL ) {
    CRIT( "Could not open file: %s\n", CONFIG_FILE );
    return -1;
  }
  
  while( !feof( cfg ) ) {
    while( fgets( line, 1024, cfg ) != NULL ) {
      if( ( !strpbrk( line, "#" ) ) && ( strlen( line ) > 1 ) ) {
	if( !strncmp( line, "LDAP_Schema: ", 13 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->ldap_schema ) ), chk + 1 );
	    if( myConfig->ldap_schema[ strlen( myConfig->ldap_schema ) - 1 ] == '\n' ) {
	      myConfig->ldap_schema[ strlen( myConfig->ldap_schema ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->ldap_servers, "NULL" );
	  }
	} else if( !strncmp( line, "LDAP_Servers: ", 13 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->ldap_servers ) ), chk + 1 );
	    if( myConfig->ldap_servers[ strlen( myConfig->ldap_servers ) - 1 ] == '\n' ) {
	      myConfig->ldap_servers[ strlen( myConfig->ldap_servers ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->ldap_servers, "NULL" );
	  }
	} else if( !strncmp( line, "LDAP_Port: ", 10 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->ldap_port ) ), chk + 1 );
	    if( myConfig->ldap_port[ strlen( myConfig->ldap_port ) - 1 ] == '\n' ) {
	      myConfig->ldap_port[ strlen( myConfig->ldap_port ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->ldap_port, "NULL" );
	  }
	} else if( !strncmp( line, "LDAP_Bind_DN: ", 14 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->ldap_bind_dn ) ), chk + 1 );
	    if( myConfig->ldap_bind_dn[ strlen( myConfig->ldap_bind_dn ) - 1 ] == '\n' ) {
	      myConfig->ldap_bind_dn[ strlen( myConfig->ldap_bind_dn ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->ldap_bind_dn, "NULL" );
	  }
	} else if( !strncmp( line, "LDAP_Base_DN: ", 14 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->ldap_base_dn ) ), chk + 1 );
	    if( myConfig->ldap_base_dn[ strlen( myConfig->ldap_base_dn ) - 1 ] == '\n' ) {
	      myConfig->ldap_base_dn[ strlen( myConfig->ldap_base_dn ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->ldap_base_dn, "NULL" );
	  }
	} else if( !strncmp( line, "LDAP_User: ", 11 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->ldap_user ) ), chk + 1 );
	    if( myConfig->ldap_user[ strlen( myConfig->ldap_user ) - 1 ] == '\n' ) {
	      myConfig->ldap_user[ strlen( myConfig->ldap_user ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->ldap_user, "NULL" );
	  }
	} else if( !strncmp( line, "LDAP_Pass: ", 11 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->ldap_pass ) ), chk + 1 );
	    if( myConfig->ldap_pass[ strlen( myConfig->ldap_pass ) - 1 ] == '\n' ) {
	      myConfig->ldap_pass[ strlen( myConfig->ldap_pass ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->ldap_pass, "NULL" );
	  }
	} else if( !strncmp( line, "LDAP_TLS: ", 10 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->ldap_tls ) ), chk + 1 );
	    if( myConfig->ldap_tls[ strlen( myConfig->ldap_tls ) - 1 ] == '\n' ) {
	      myConfig->ldap_tls[ strlen( myConfig->ldap_tls ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->ldap_tls, "NULL" );
	  }
	} else if( !strncmp( line, "MYSQL_Servers: ", 15 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->mysql_servers ) ), chk + 1 );
	    if( myConfig->mysql_servers[ strlen( myConfig->mysql_servers ) - 1 ] == '\n' ) {
	      myConfig->mysql_servers[ strlen( myConfig->mysql_servers ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->mysql_servers, "NULL" );
	  }
	} else if( !strncmp( line, "MYSQL_Username: ", 16 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->mysql_username ) ), chk + 1 );
	    if( myConfig->mysql_username[ strlen( myConfig->mysql_username ) - 1 ] == '\n' ) {
	      myConfig->mysql_username[ strlen( myConfig->mysql_username ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->mysql_username, "NULL" );
	  }
	} else if( !strncmp( line, "MYSQL_Password: ", 16 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->mysql_password ) ), chk + 1 );
	    if( myConfig->mysql_password[ strlen( myConfig->mysql_password ) - 1 ] == '\n' ) {
	      myConfig->mysql_password[ strlen( myConfig->mysql_password ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->mysql_password, "NULL" );
	  }
	} else if( !strncmp( line, "MYSQL_Database: ", 16 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->mysql_database ) ), chk + 1 );
	    if( myConfig->mysql_database[ strlen( myConfig->mysql_database ) - 1 ] == '\n' ) {
	      myConfig->mysql_database[ strlen( myConfig->mysql_database ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->mysql_database, "NULL" );
	  }
	} else if( !strncmp( line, "MYSQL_SSL: ", 11 ) ) {
	  chk = strpbrk( line, "  " );
	  if( chk != '\0' ) {
	    strcpy( *( &( myConfig->mysql_ssl ) ), chk + 1 );
	    if( myConfig->mysql_ssl[ strlen( myConfig->mysql_ssl ) - 1 ] == '\n' ) {
	      myConfig->mysql_ssl[ strlen( myConfig->mysql_ssl ) - 1 ] = '\0';
	    }
	  } else {
	    strcpy( myConfig->mysql_ssl, "NULL" );
	  }
	} else {
	  return 1;
	}
      }
    }
  }
  fflush( cfg );
  fclose( cfg );
  return 0;
}
#endif

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags,
                                   int argc, const char **argv) {
  int pamRet;
  const char *user;
  int tries;
  
  CRIT( "Begining pam_libdpfp authentication..." );

  pamRet = pam_fail_delay(pamh, 100);
  
  pamRet = pam_get_user(pamh, &user, NULL);
  if (pamRet != PAM_SUCCESS || strlen(user) == 0) {
    if (pamRet == PAM_CONV_AGAIN) {
      pamRet = PAM_INCOMPLETE;
    } else {
      CRIT( "Error retrieving user name: %s", pam_strerror( pamh, pamRet ) );
      pamRet = PAM_USER_UNKNOWN;
    }
    closelog();
    return pamRet;
  }

  dpfp_aquire_fp();
  
#ifdef INCLUDE_DATABASES
  if( ReadConfig() != -1 ) {
    return;
  } else {
    CRIT( "Error opening configuration file - pam_libdpfp.conf" );
  }
#endif
  
#ifdef HAVE_LIBLDAP
  DEBUG( "libldap enabled" );
  if( LDAPChkOptions() == -1 ) {
    CRIT( "Error with configuation data for OpenLDAP/Active Directory - pam_libdpfp.conf" );
  } else {
    if( CheckLDAP() == -1 ) {
      CRIT( "Error with user lookup in OpenLDAP/Active Directory" );
    }
  }
#endif
  
#ifdef HAVE_LIBMYSQL
  DEBUG( "libmysql enabled" );
#endif

#ifdef HAVE_LIBKRB5
  DEBUG( "libkrb5 enabled" );
#endif

}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags,
			      int argc, const char **argv) {
  return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv) {
  int pamRet;
  const char *user;	
  const char *uuidString;
  const char *birDbPath;
  int tries;
  
  pamRet = pam_fail_delay(pamh, 100);
  if ((flags & PAM_PRELIM_CHECK) == PAM_PRELIM_CHECK) {
    return PAM_SUCCESS; /* No remote resource to check */
  }
  
  pamRet = pam_get_user(pamh, &user, NULL);
  if (pamRet != PAM_SUCCESS || strlen(user) == 0) {
    if (pamRet == PAM_CONV_AGAIN) {
      pamRet = PAM_INCOMPLETE;
    } else {
      pamRet = PAM_USER_UNKNOWN;
    }
    return pamRet;
  }
}

int dpfp_aquire_fp(void)
{
  int result;
  struct dpfp_dev *dev;
  struct dpfp_fprint *base_img = dpfp_fprint_alloc();
  struct dpfp_fprint *fp = dpfp_fprint_alloc();
  struct dpfp_fprint *mask = dpfp_fprint_alloc();
  struct dpfp_ffield *direction = dpfp_ffield_alloc();
  struct dpfp_ffield *frequency = dpfp_ffield_alloc();
  struct timeval tv;
  double t1, t2;
  
  dpfp_init();
  dev = dpfp_open();

  if (dev == NULL) {
    perror("dev");
    goto exit;
  }
  
  /* Capture base image. We assume the finger is away from the sensor at
   * this point... */
  if (dpfp_set_mode(dev, DPFP_MODE_SEND_FINGER) < 0) {
    CRIT("set_mode for libdpfp failed");
    goto exit;
  }
  
  if (dpfp_capture_fprint(dev, base_img) < 0) {
    CRIT("simple_get_fingerprint in libdpfp failed");
    goto exit;
  }
  
  if (dpfp_simple_await_finger_on(dev) < 0) {
    CRIT("await_finger_on in libdpfp failed");
    goto exit;
  }
  
  /* Capture fingerprint */
  if (dpfp_set_mode(dev, DPFP_MODE_SEND_FINGER) < 0) {
    CRIT("set_mode in libdpfp failed");
    goto exit;
  }
  
  if (dpfp_capture_fprint(dev, fp) < 0) {
    perror("simple_get_fingerprint");
    goto exit;
  }
  
  dpfp_close(dev);

  gettimeofday(&tv, NULL);
  t1 = TV_TO_DOUBLE(tv);

  /* Basic enhancements: subtract base image, flip to correct orientation */
  dpfp_fprint_subtract(fp, base_img);
  dpfp_fprint_flip_v(fp);
  dpfp_fprint_flip_h(fp);

  if (dpfp_fprint_write_to_file(fp, "finger.pgm") < 0)
    perror("write_fingerprint_to_file");

  /* More advanced enhancements */
  dpfp_fprint_soften_mean(fp, 3);
  dpfp_fprint_get_direction(fp, direction, 7, 8);
  dpfp_fprint_get_frequency(fp, direction, frequency);
  dpfp_fprint_get_mask(fp, direction, frequency, mask);
  dpfp_fprint_enhance_gabor(fp, direction, frequency, mask, 4.0);
  dpfp_fprint_binarize(fp, 0x80);
  
  if (dpfp_fprint_write_to_file(fp, "finger_enhanced.pgm") < 0)
    CRIT("write_fingerprint_to_file in libdpfp failed");
  
  dpfp_fprint_thin(fp);
  
  gettimeofday(&tv, NULL);
  t2 = TV_TO_DOUBLE(tv);
  DEBUG("enhancements took %.6lf seconds in total\n", t2 - t1);
  
  if (dpfp_fprint_write_to_file(fp, "finger_thinned.pgm") < 0)
    CRIT("write_fingerprint_to_file in libdpfp failed");
  
  return 0;
  
 exit:
  dpfp_close(dev);
  return 1;
}

#ifdef INCLUDE_LIBLDAP
static int LDAPChkOptions()
{
  myConfigPtr myConfig = &localConfig;
  
  if( ( myConfig->ldap_schema[ strlen( myConfig->ldap_schema ) -1 ] != '\0' ) &&
      ( ( myConfig->ldap_schema != "LDAP" ) || ( myConfig->ldap_schema != "AD" ) ) &&
      ( myConfig->ldap_servers[ strlen( myConfig->ldap_servers ) -1 ] != '\0' ) &&
      ( myConfig->ldap_user[ strlen( myConfig->ldap_user ) -1 ] != '\0' ) &&
      ( myConfig->ldap_pass[ strlen( myConfig->ldap_pass ) -1 ] != '\0' ) &&
      ( myConfig->ldap_base_dn[ strlen( myConfig->ldap_base_dn ) -1 ] != '\0' ) ) {
    return 0;
  } else {
    return -1;
  }
}

int CheckLDAP( int argsv, const char *argv[] )
{
  int ret;
  int rdConfig = 0;
  
  LDAP            *ld;
  LDAPMessage     *res, *e;
  int             i, rc;
  char            *attrs[5];
  char            buffer[80];
  char            *host;
  char            *base;
  char            **vals;
  int             port;
  const char      *user;
  const char      *pass;
  BerElement      **berptr;
  BerElement      * ber;
  
  myConfigPtr myConfig = &localConfig;
  berptr = &ber;
  
  attrs[ 0 ] = "cn";
  attrs[ 1 ] = "uidNumber";
  attrs[ 2 ] = "gidNumber";
  attrs[ 3 ] = "loginShell";
  
  /* check to see if were looking at AD or LDAP */
  if( myConfig->ldap_schema == "AD" ) {
    attrs[ 4 ] = "msSFUHomeDirectory";
  } else {
    attrs[ 4 ] = "homedirectory";
  }
  
  if( ( rdConfig = ReadConfig() ) != 0 ) {
    CRIT( "Unable to open %s for database options", CONFIG_FILE );
    ret = 1;
  } else {
    if( LDAPChkOptions() == 1 ) {
      CRIT( "No LDAP / Active Directory options defined in %s", CONFIG_FILE );
      ret = 1;
    } else {
      if( myConfig->ldap_servers != '\0' ) {
	host = myConfig->ldap_servers;
      }
      
      if( myConfig->ldap_port != '\0' ) {
	port = atoi( myConfig->ldap_port );
      }
      
      if( myConfig->ldap_user != '\0' ) {
	user = myConfig->ldap_user;
      }
      
      if( myConfig->ldap_pass != '\0' ) {
	pass = myConfig->ldap_pass;
      }
      
      if( myConfig->ldap_base_dn != '\0' ) {
	base = myConfig->ldap_base_dn;
      }
      
      if( ( ld = ldap_init( host, port ) ) == NULL ) {
	CRIT( "Initialization Failed: %s", host );
	return 1;
      }

      if( myConfig->ldap_tls == "true" ) {
	/* fix this
	rc = ldap_sasl_bind_s( ld, user, pass, NULL, NULL, NULL, NULL );
	rc = ldap_parse_sasl_bind_result( ld, rc, rc, 1 );
	*/
      } else {
	rc = ldap_simple_bind_s( ld, user, pass );
      }
      
      switch( rc ) {
      case LDAP_SUCCESS:
	
	sprintf( buffer, "(cn=%s)", myConfig->usrname );
	notice( "Searching for: %s", buffer );
	
	if ( ldap_search_s( ld, base, LDAP_SCOPE_SUBTREE,
			    buffer, NULL, 0, &res ) != LDAP_SUCCESS ) {
	  warn( "No results for %s", buffer );
	  return 1;
	}
	
	if( ldap_first_entry( ld, res ) == NULL ) {
	  warn( "%s not found in LDAP/AD aborting authentication", buffer );
	}
	
	for( e = ldap_first_entry( ld, res ); e != NULL; e = ldap_next_entry( ld, e )) {
	  if(( vals = ldap_get_values( ld, e, "cn" )) != NULL ) {
	    for( i = 0; vals[i] != NULL; i++ ) {
	      strcpy( myConfig->id, vals[i] );
	    }
	  }
	  if(( vals = ldap_get_values( ld, e, "uidNumber" )) != NULL ) {
	    for( i = 0; vals[i] != NULL; i++ ) {
	      strcpy( myConfig->uid, vals[i] );
	    }
	  }
	  if(( vals = ldap_get_values( ld, e, "gidNumber" )) != NULL ) {
	    for( i = 0; vals[i] != NULL; i++ ) {
	      strcpy( myConfig->gid, vals[i] );
	    }
	  }
	  if(( vals = ldap_get_values( ld, e, "loginShell" )) != NULL ) {
	    for( i = 0; vals[i] != NULL; i++ ) {
	      if( myConfig->shell == NULL ) {
		strcpy( myConfig->shell, vals[i] );
	      }
	    }
	  }
	  if(( vals = ldap_get_values( ld, e, "msSFUHomeDirectory" )) != NULL ) {
	    for( i = 0; vals[i] != NULL; i++ ) {
	      if( myConfig->home == NULL ) {
		strcpy( myConfig->home, vals[i] );
	      }
	    }
	  }
	}
	ldap_msgfree( res );
	ldap_unbind( ld );
	return 0;
	
      default:
	CRIT( "Error reported: %s\n", myConfig->ldap_servers );
	return 1;
      }
      return 0;
    }
  }
}
#endif

#ifdef INCLUDE_LIBMYSQL
int CheckMysql()
{
  MYSQL *conn;
  MYSQL_RES *res;
  MYSQL_ROW row;
  
  int x, result = 0;
  myConfigPtr myConfig = &localConfig;
  
  /* convert vars */
  char *server = myConfig->mysql_servers;
  char *user = myConfig->mysql_username;
  char *pass = myConfig->mysql_password;
  char *db = myConfig->mysql_database;
  
  if( ( x = ReadConfig() ) != 0 ) {
    CRIT( "Unable to open %s for database options", CONFIG_FILE );
    result = 1;
  } else {
    conn = mysql_init( NULL );
    if( !mysql_real_connect( conn, server, user, pass, db, 0, NULL, 0 ) ) {
      CRIT( "Unable to connect to MySQL database" );
      result = 1;
    }
    if( mysql_query( conn, "SELECT * FROM auth WHERE bin = " ) ) {
      CRIT( "Unable to find user matching" );
      result = 1;
    }
    res = mysql_use_result( conn );
    if( result != 1 ) {
      while( ( row = mysql_fetch_row( res ) ) != NULL ) {
	CRIT( "%s was found in database", row[1] );
      }
    }
    mysql_free_result( res );
    mysql_close( conn );
    return result;
  }
}
#endif

#ifdef INCLUDE_LIBKRB5
/* Sections shamelessly ripped from pam_krb5 */
int
_pam_krb5_conv_call(pam_handle_t *pamh,
		    PAM_KRB5_MAYBE_CONST struct pam_message *messages,
		    int n_prompts,
		    struct pam_response **responses)
{
	struct pam_conv *conv;
	int i;
	struct pam_response *drop_responses;
	PAM_KRB5_MAYBE_CONST struct pam_message **message_array;

	/* Get the address of the conversation structure provided by the
	 * application. */
	conv = NULL;
	i = _pam_krb5_get_item_conv(pamh, &conv);
	if (i != PAM_SUCCESS) {
		return i;
	}
	if (conv == NULL) {
		return PAM_BAD_ITEM;
	}

	/* Allocate the array for storing pointers to elements in the array
	 * which we were passed. */
	message_array = malloc(sizeof(struct pam_message*) * n_prompts);
	if (message_array == NULL) {
		return PAM_BUF_ERR;
	}
	memset(message_array, 0, sizeof(struct pam_message*) * n_prompts);

	/* Initialize the array so that each element holds a pointer to a
	 * corresponding element in the passed-in array. */
	for (i = 0; i < n_prompts; i++) {
		message_array[i] = &(messages[i]);
	}

	/* Call the converation function. */
	if (responses == NULL) {
		responses = &drop_responses;
		drop_responses = NULL;
	}
	i = conv->conv(n_prompts, message_array, responses, conv->appdata_ptr);
	if (responses == &drop_responses) {
		_pam_krb5_maybe_free_responses(drop_responses, n_prompts);
	}

	/* Clean up. */
	memset(message_array, 0, sizeof(struct pam_message*) * n_prompts);
	free(message_array);

	return i;
}

int
krb5_authenticate(pam_handle_t *pamh, int flags,
		    int argc, PAM_KRB5_MAYBE_CONST char **argv)
{
	PAM_KRB5_MAYBE_CONST char *user;
	krb5_context ctx;
	struct _pam_krb5_options *options;
	struct _pam_krb5_user_info *userinfo;
	struct _pam_krb5_stash *stash;
	krb5_get_init_creds_opt gic_options;
	int i, retval;
	char *password;

	/* Initialize Kerberos. */
	if (_pam_krb5_init_ctx(&ctx, argc, argv) != 0) {
		warn("error initializing Kerberos");
		return PAM_SERVICE_ERR;
	}

	/* Get the user's name. */
	i = pam_get_user(pamh, &user, NULL);
	if (i != PAM_SUCCESS) {
		warn("could not identify user name");
		krb5_free_context(ctx);
		return i;
	}

	/* Read our options. */
	options = _pam_krb5_options_init(pamh, argc, argv, ctx);
	if (options == NULL) {
		warn("error parsing options (shouldn't happen)");
		krb5_free_context(ctx);
		return PAM_SERVICE_ERR;
	}
	if (options->debug) {
		debug("called to authenticate '%s', realm '%s'", user,
		      options->realm);
	}
	krb5_get_init_creds_opt_init(&gic_options);
	_pam_krb5_set_init_opts(ctx, &gic_options, options);

	/* Get information about the user and the user's principal name. */
	userinfo = _pam_krb5_user_info_init(ctx, user, options->realm,
					    options->user_check,
					    options->n_mappings,
					    options->mappings);
	if (userinfo == NULL) {
		if (options->ignore_unknown_principals) {
			retval = PAM_IGNORE;
		} else {
			warn("error getting information about '%s'", user);
			retval = PAM_USER_UNKNOWN;
		}
		if (options->use_second_pass) {
			password = NULL;
			i = _pam_krb5_prompt_for(pamh, "Password: ", &password);
			if ((i == PAM_SUCCESS) &&
			    (flags & PAM_DISALLOW_NULL_AUTHTOK) &&
			    (strlen(password) == 0)) {
				warn("disallowing NULL authtok for '%s'", user);
				retval = PAM_AUTH_ERR;
				i = PAM_AUTH_ERR;
			}
			if (i == PAM_SUCCESS) {
				/* Save the password for the next module. */
				if (!_pam_krb5_has_item(pamh, PAM_AUTHTOK)) {
					if (options->debug) {
						debug("saving newly-entered "
						      "password for use by "
						      "other modules");
					}
					pam_set_item(pamh, PAM_AUTHTOK, password);
				}
			}
		}
		_pam_krb5_options_free(pamh, ctx, options);
		krb5_free_context(ctx);
		return retval;
	}
	if (options->debug) {
		debug("authenticating '%s'", userinfo->unparsed_name);
	}

	/* Check the minimum UID argument. */
	if ((options->minimum_uid != (uid_t) -1) &&
	    (userinfo->uid < options->minimum_uid)) {
		if (options->debug) {
			debug("ignoring '%s' -- uid below minimum = %lu", user,
			      (unsigned long) options->minimum_uid);
		}
		_pam_krb5_user_info_free(ctx, userinfo);
		_pam_krb5_options_free(pamh, ctx, options);
		krb5_free_context(ctx);
		return PAM_IGNORE;
	}

	/* Get the stash for this user. */
	stash = _pam_krb5_stash_get(pamh, userinfo, options);
	if (stash == NULL) {
		warn("error retrieving stash for '%s' (shouldn't happen)",
		     user);
		_pam_krb5_user_info_free(ctx, userinfo);
		_pam_krb5_options_free(pamh, ctx, options);
		krb5_free_context(ctx);
		return PAM_SERVICE_ERR;
	}

	/* If we've been called before, then the stash is more or less stale,
	 * so reset things for applications which call pam_authenticate() more
	 * than once with the same library context. */
	stash->v5attempted = 0;

	retval = PAM_AUTH_ERR;

	/* If we're configured to use an existing ccache, try that. */
	if ((retval != PAM_SUCCESS) && (options->existing_ticket)) {
		if (options->debug) {
			debug("trying existing credentials for '%s'",user);
		}
		retval = v5_get_creds(ctx, pamh,
				      &stash->v5creds, userinfo,
				      options,
				      KRB5_TGS_NAME,
				      password, &gic_options,
				      &stash->v5result);
		stash->v5attempted = 1;
		if (options->debug) {
			debug("got result %d (%s)", stash->v5result,
			      v5_error_message(stash->v5result));
		}
	}

	/* Try with the stored password, if we've been told to do so. */
	if ((retval != PAM_SUCCESS) && (options->use_first_pass)) {
		password = NULL;
		i = _pam_krb5_get_item_text(pamh, PAM_AUTHTOK, &password);
		if ((i == PAM_SUCCESS) &&
		    (flags & PAM_DISALLOW_NULL_AUTHTOK) &&
		    (password != NULL) &&
		    (strlen(password) == 0)) {
			warn("disallowing NULL authtok for '%s'", user);
			retval = PAM_AUTH_ERR;
			i = PAM_AUTH_ERR;
		}
		if ((i == PAM_SUCCESS) && (password != NULL)) {
			if (options->debug) {
				debug("trying previously-entered password for "
				      "'%s'", user);
			}
			retval = v5_get_creds(ctx, pamh,
					      &stash->v5creds, userinfo,
					      options,
					      KRB5_TGS_NAME,
					      password, &gic_options,
					      &stash->v5result);
			stash->v5attempted = 1;
			if (options->debug) {
				debug("got result %d (%s)", stash->v5result,
				      v5_error_message(stash->v5result));
			}
		}
		if ((retval == PAM_SUCCESS) &&
		    ((options->v4 == 1) || (options->v4_for_afs == 1))) {
			v4_get_creds(ctx, pamh, stash, userinfo, options,
				     password, &i);
			if ((i != 0) && (options->debug)) {
				debug("error obtaining v4 creds: %d (%s)",
				      i, v5_error_message(i));
			}
			if (stash->v4present &&
			    (options->ignore_afs == 0) &&
			    (options->tokens == 1)) {
				v5_save(ctx, stash, userinfo, options, NULL);
				v4_save(ctx, stash, userinfo, options,
					getuid(), getgid(), NULL);
				tokens_obtain(ctx, stash, options, userinfo, 1);
				v4_destroy(ctx, stash, options);
				v5_destroy(ctx, stash, options);
			}
		}
	}

	/* If that didn't work, ask for a new password and try again. */
	if ((retval != PAM_SUCCESS) && (options->use_second_pass)) {
		password = NULL;
		i = _pam_krb5_prompt_for(pamh, "Password: ", &password);
		if ((i == PAM_SUCCESS) &&
		    (flags & PAM_DISALLOW_NULL_AUTHTOK) &&
		    (password != NULL) &&
		    (strlen(password) == 0)) {
			warn("disallowing NULL authtok for '%s'", user);
			retval = PAM_AUTH_ERR;
			i = PAM_AUTH_ERR;
		}
		if ((i == PAM_SUCCESS) && (password != NULL)) {
			/* Save the password for the next module. */
			if (!_pam_krb5_has_item(pamh, PAM_AUTHTOK)) {
				if (options->debug) {
					debug("saving newly-entered "
					      "password for use by "
					      "other modules");
				}
				pam_set_item(pamh, PAM_AUTHTOK, password);
			}
			/* Get creds. */
			if (options->debug) {
				debug("trying newly-entered password for "
				      "'%s'", user);
			}
			retval = v5_get_creds(ctx, pamh,
					      &stash->v5creds, userinfo,
					      options,
					      KRB5_TGS_NAME,
					      password, &gic_options,
					      &stash->v5result);
			stash->v5attempted = 1;
			if (options->debug) {
				debug("got result %d (%s)", stash->v5result,
				      v5_error_message(stash->v5result));
			}
			/* Save the password for the next module. */
			if (!_pam_krb5_has_item(pamh, PAM_AUTHTOK)) {
				if (options->debug) {
					debug("saving newly-entered "
					      "password for use by "
					      "other modules");
				}
				pam_set_item(pamh, PAM_AUTHTOK, password);
			}
		} else {
			warn("error reading password for '%s'", user);
		}
		if ((retval == PAM_SUCCESS) &&
		    ((options->v4 == 1) || (options->v4_for_afs == 1))) {
			v4_get_creds(ctx, pamh, stash, userinfo, options,
				     password, &i);
			if ((i != 0) && (options->debug)) {
				debug("error obtaining v4 creds: %d (%s)",
				      i, v5_error_message(i));
			}
			if (stash->v4present &&
			    (options->ignore_afs == 0) &&
			    (options->tokens == 1)) {
				v5_save(ctx, stash, userinfo, options, NULL);
				v4_save(ctx, stash, userinfo, options,
					getuid(), getgid(), NULL);
				tokens_obtain(ctx, stash, options, userinfo, 1);
				v4_destroy(ctx, stash, options);
				v5_destroy(ctx, stash, options);
			}
		}
		if (password != NULL) {
			xstrfree(password);
		}
	}

	/* If we didn't use any password, try here. */
	if ((retval != PAM_SUCCESS) &&
	    (!options->use_first_pass) && (!options->use_second_pass)) {
		password = NULL;
		retval = v5_get_creds(ctx, pamh,
				      &stash->v5creds, userinfo,
				      options,
				      KRB5_TGS_NAME,
				      NULL, &gic_options,
				      &stash->v5result);
		stash->v5attempted = 1;
		if (options->debug) {
		  debug( "4" );
			debug("got result %d (%s)", stash->v5result,
			      v5_error_message(stash->v5result));
		}
		if ((retval == PAM_SUCCESS) &&
		    ((options->v4 == 1) || (options->v4_for_afs == 1))) {
			v4_get_creds(ctx, pamh, stash, userinfo, options,
				     password, &i);
			if ((i != 0) && (options->debug)) {
				debug("error obtaining v4 creds: %d (%s)",
				      i, v5_error_message(i));
			}
			if (stash->v4present &&
			    (options->ignore_afs == 0) &&
			    (options->tokens == 1)) {
				v5_save(ctx, stash, userinfo, options, NULL);
				v4_save(ctx, stash, userinfo, options,
					getuid(), getgid(), NULL);
				tokens_obtain(ctx, stash, options, userinfo, 1);
				v4_destroy(ctx, stash, options);
				v5_destroy(ctx, stash, options);
			}
		}
	}

	/* If we got this far, check the target user's .k5login file. */
	if ((retval == PAM_SUCCESS) && options->user_check) {
		if (krb5_kuserok(ctx, userinfo->principal_name, user) == 0) {
			notice("account checks fail for '%s': user disallowed "
			       "by .k5login file for '%s'",
			       userinfo->unparsed_name, user);
			retval = PAM_PERM_DENIED;
		}
	}

	/* Log the authentication status, optionally saving the credentials in
	 * a piece of shared memory. */
	if (retval == PAM_SUCCESS) {
		if (options->use_shmem) {
			_pam_krb5_stash_shm_write(pamh, stash, options,
						  userinfo);
		}
		notice("authentication succeeds for '%s' (%s)", user,
		       userinfo->unparsed_name);
	} else {
		if ((retval == PAM_USER_UNKNOWN) &&
		    options->ignore_unknown_principals) {
			retval = PAM_IGNORE;
		} else {
			notice("authentication fails for '%s' (%s): %s (%s)",
			       user,
			       userinfo->unparsed_name,
			       pam_strerror(pamh, retval),
			       v5_error_message(stash->v5result));
		}
	}
	/* Clean up. */
	if (options->debug) {
		debug("pam_authenticate returning %d (%s)", retval,
		      pam_strerror(pamh, retval));
	}
	_pam_krb5_options_free(pamh, ctx, options);
	_pam_krb5_user_info_free(ctx, userinfo);
	krb5_free_context(ctx);
	return retval;
}

int
krb5_setcred(pam_handle_t *pamh, int flags,
	       int argc, PAM_KRB5_MAYBE_CONST char **argv)
{
	if (flags & PAM_ESTABLISH_CRED) {
		return pam_sm_open_session(pamh, flags, argc, argv);
	}
	if (flags & (PAM_REINITIALIZE_CRED | PAM_REFRESH_CRED)) {
		if (_pam_krb5_sly_looks_unsafe() == 0) {
			return _pam_krb5_sly_maybe_refresh(pamh, flags,
							   argc, argv);
		} else {
			return PAM_IGNORE;
		}
	}
	if (flags & PAM_DELETE_CRED) {
		return pam_sm_close_session(pamh, flags, argc, argv);
	}
	warn("pam_setcred() called with no flags");
	return PAM_SERVICE_ERR;
}
#endif

